/* IR-to-BLE gateway application for RFDuino
 * Gabor Paller, 2015
 */

#include <core_cm0.h>
#include <nrf_gpiote.h>
#include <nrf51.h>
#include <IRLib.h>
#include <IRLibRData.h>
#include <RFduinoBLE.h>
#include <Wire2.h>

int SDA_PIN = 2;
int SCL_PIN = 3;
int INT_PIN = 4;
int RXD_PIN = 5;
int TXD_PIN = 6;
int MCP23008_ADDR = 0x20;
int MCP23008_REG_IODIR = 0x00;
int MCP23008_REG_IPOL  = 0x01;
int MCP23008_REG_GPINTEN  = 0x02;
int MCP23008_REG_DEFVAL  = 0x03;
int MCP23008_REG_INTCON  = 0x04;
int MCP23008_REG_IOCON = 0x05;
int MCP23008_REG_GPPU = 0x06;
int MCP23008_REG_INTF = 0x07;
int MCP23008_REG_INTCAP = 0x08;
int MCP23008_REG_GPIO =  0x09;
int MCP23008_REG_OLAT =  0x0A;

int clientConnected = 0;
extern volatile irparams_t irparams;
volatile int prev_intcap;
uint8_t twi_send_buf[1];
uint8_t twi_recv_buf[1];

void writeReg(int reg,int value) {
  Wire2.beginTransmission(MCP23008_ADDR); // transmit to the MCP23008
  Wire2.write((byte)reg); 
  Wire2.write((byte)value);
  Wire2.endTransmission();
}

int readReg(int reg) {
  int b;
  Wire2.beginTransmission(MCP23008_ADDR); // transmit to the MCP23008
  Wire2.write(byte(MCP23008_REG_INTCAP)); 
  Wire2.endTransmission();
  Wire2.requestFrom(MCP23008_ADDR,1 );
  b = Wire2.read();
  return b;
}

class IRrecvRFDuinoLoop: public IRrecvBase
{
public:
  IRrecvRFDuinoLoop():IRrecvBase(0){};
  bool GetResults(IRdecodeBase *decoder);
};

bool IRrecvRFDuinoLoop::GetResults(IRdecodeBase *decoder) {
  bool Finished=false;
  byte OldState;
  byte NewState;
  int prev_intcap,intcap;
  int delta_intcap;
  uint32_t StartTime, DeltaTime, EndTime;

  StartTime=micros();
  OldState = digitalRead( INT_PIN );
  prev_intcap = readReg( MCP23008_REG_INTCAP ) & 0x7F;
  while(irparams.rawlen<RAWBUF) {  //While the buffer not overflowing
    while( 1 ) {
	EndTime = micros();
	DeltaTime = EndTime - StartTime;
	NewState = digitalRead(INT_PIN);
	if( ( OldState == LOW ) && ( NewState == HIGH ) )
		break;
	OldState = NewState;
      	if( DeltaTime > 10000) { //If it's a very long wait
        	if(Finished = irparams.rawlen) 
			break; //finished unless it's the opening gap
      	}
    }
    if(Finished) break;
// We look for 0->1 transitions
    if( NewState == HIGH ) {
      do_Blink();
// With our clock circuit, the maximum measuring time is approximately 5000 microsec. Use the micros() if the
// time is very long. But read INTCAP anyway to clear the interrupt
      intcap = readReg( MCP23008_REG_INTCAP ) & 0x7F;
      OldState = digitalRead(INT_PIN);
      StartTime=EndTime;
      if( DeltaTime > 5000 )
      	irparams.rawbuf[irparams.rawlen++]= DeltaTime;
      else {
        delta_intcap = intcap - prev_intcap;
        if( delta_intcap < 0 )
		delta_intcap += 128;
        prev_intcap = intcap;
// main oscillator at 350 kHz, prescaler of 16 yields approx. 22 kHz clock, 43 microsec cycle but matcher calculates
// with half cycle time->22
        irparams.rawbuf[irparams.rawlen++]= (unsigned int)delta_intcap * 22;
      }
    }
  }
  decoder->Reset();//clear out any old values.
  decoder->rawlen = irparams.rawlen;
  return true;
}

class IRrecvRFDuinoPCI: public IRrecvBase
{
public:
  IRrecvRFDuinoPCI();
  bool GetResults(IRdecodeBase *decoder);
  void resume(void);
};

void twi_complete(int recvbytes) {
  unsigned long volatile ChangeTime;
  unsigned long DeltaTime;
  int intcap,delta_intcap;
  bool store;
  
  if( recvbytes > 0 ) {
    ChangeTime=micros();
    DeltaTime = ChangeTime-irparams.timer;
    intcap = ((int)twi_recv_buf[0]) & 0x7F;	// MCP23008_REG_INTCAP, delivered by the interrupt routine
    if( DeltaTime <= 5000 ) {
      if( prev_intcap >= 0 ) {
        delta_intcap = intcap - prev_intcap;
        if( delta_intcap < 0 )
		delta_intcap += 128;
// main oscillator at 350 kHz, prescaler of 16 yields approx. 22 kHz clock, 43 microsec cycle but matcher calculates
// with half cycle time->22
        DeltaTime = (unsigned int)delta_intcap * 22;
      }
    }
    prev_intcap = intcap;
    store = true;
    switch(irparams.rcvstate) {
	case STATE_STOP:
		store = false; 
		break;

	case STATE_RUNNING:
        	if( DeltaTime>10000 ) {
          	  irparams.rcvstate = STATE_STOP; 
        //Setting gap to 0 is a flag to let you know why we stopped For debugging purposes
        //irparams.rawbuf[0]=0;
	  	  store = false;
        	}
        	break;

	case STATE_IDLE:
		irparams.rcvstate=STATE_RUNNING;
       		break;

    	case STATE_UNKNOWN:
    	case STATE_MARK:
    	case STATE_SPACE:
       		break;
    }
    if( store ) {
    	irparams.rawbuf[irparams.rawlen]=DeltaTime;
        if(++irparams.rawlen>=RAWBUF) {
          irparams.rcvstate=STATE_STOP;
    //Setting gap to 1 is a flag to let you know why we stopped For debugging purposes
    //irparams.rawbuf[0]=1;
        }
    }
    irparams.timer=ChangeTime;
  } 
}

void GPIOTE_Interrupt(void){ 
  if( (NRF_GPIOTE->EVENTS_IN[0] == 1) && (NRF_GPIOTE->INTENSET & GPIOTE_INTENSET_IN0_Msk))
	NRF_GPIOTE->EVENTS_IN[0] = 0;
// Start the reading of the capture register. This time the process will only be started, the entire process
// is handled in interrupts
// With our clock circuit, the maximum measuring time is approximately 5000 microsec. Use the micros() if the
// time is very long. But read INTCAP anyway to clear the interrupt
  twi_send_buf[0] = MCP23008_REG_INTCAP;
  Wire2.sendReceiveInt(MCP23008_ADDR, 1, twi_send_buf, 1, twi_recv_buf);
}


IRrecvRFDuinoPCI::IRrecvRFDuinoPCI() {
  Init();
}


void IRrecvRFDuinoPCI::resume(void) {
  irparams.rcvstate = STATE_IDLE;
  IRrecvBase::resume();
  irparams.timer=micros();
  pinMode(INT_PIN, INPUT); 
  NRF_GPIOTE->CONFIG[0] =  (GPIOTE_CONFIG_POLARITY_LoToHi << GPIOTE_CONFIG_POLARITY_Pos)
              | (INT_PIN << GPIOTE_CONFIG_PSEL_Pos) 
              | (GPIOTE_CONFIG_MODE_Event << GPIOTE_CONFIG_MODE_Pos);
  NRF_GPIOTE->INTENSET = GPIOTE_INTENSET_IN0_Set << GPIOTE_INTENSET_IN0_Pos;
  Wire2.onReceive(twi_complete);
  attachInterrupt(GPIOTE_IRQn, GPIOTE_Interrupt);
// Clear the interrupt before we start serving interrupts    
  readReg( MCP23008_REG_INTCAP );
  Wire2.enableIT(); 
}

bool IRrecvRFDuinoPCI::GetResults(IRdecodeBase *decoder) {
  if(irparams.rcvstate==STATE_RUNNING) {
    unsigned long ChangeTime=irparams.timer;
    if( (micros()-ChangeTime) > 10000) {
      irparams.rcvstate=STATE_STOP;
      //Setting gap to 2 is a flag to let you know why we stopped For debugging purposes
      //irparams.rawbuf[0]=2;
    } else
	RFduino_ULPDelay(200);
  }
  if (irparams.rcvstate != STATE_STOP) return false;
  detachInterrupt(GPIOTE_IRQn);
  Wire2.disableIT();
  IRrecvBase::GetResults(decoder);
  return true;
}



IRrecvRFDuinoPCI My_Receiver;

IRdecode My_Decoder;
uint32_t Buffer[RAWBUF];

void setup()
{
// Kick the BLE stack
  RFduinoBLE.advertisementInterval = 675;
  RFduinoBLE.deviceName = "irblegw";
  RFduinoBLE.begin();
// Starts the UART at 9600bps with RX on GPIO4 and TX on GPIO5
  Serial.begin(9600,RXD_PIN,TXD_PIN);
// Sets up the MCP23008 interrupt service
  pinMode(INT_PIN, INPUT); 

// Sets up the MC23008
  clientConnected = 0;
// We do need that 400kHz I2C clock, in the main loop we can't waste time
  Wire2.speed = 400;
  Wire2.beginOnPins(SCL_PIN, SDA_PIN);
  writeReg(MCP23008_REG_IODIR,0xFF);
  writeReg(MCP23008_REG_GPPU,0x00);
  writeReg(MCP23008_REG_IPOL,0x00);
  writeReg(MCP23008_REG_INTCON,0x00);
  writeReg(MCP23008_REG_IOCON,0x02);
  writeReg(MCP23008_REG_GPINTEN,0x80);
  prev_intcap = -1;
  My_Receiver.resume();
  Serial.println("ready");
}

void loop() {
  unsigned char data[4];
  bool status = false;
  bool NewState;

  status = My_Receiver.GetResults(&My_Decoder);
  if (status) {
    //Restart the receiver so it can be capturing another code
    //while we are working on decoding this one.
    My_Receiver.resume(); 
    My_Decoder.decode();
// Don't care about REPEAT code
    if( clientConnected ) {
// If decoding was unsuccessful or key code is REPEAT, don't send
	if( ( !( ( My_Decoder.decode_type == UNKNOWN ) || ( My_Decoder.decode_type > LAST_PROTOCOL ) ) )  
		&& !( My_Decoder.value == REPEAT ) ) {
		data[0] = (unsigned char)( My_Decoder.value & 0xFFL );
		data[1] = (unsigned char)( ( My_Decoder.value & 0xFF00L ) >> 8 );
		data[2] = (unsigned char)( ( My_Decoder.value & 0xFF0000L ) >> 16 );
		data[3] = (unsigned char)( ( My_Decoder.value & 0xFF000000L ) >> 24 );
		RFduinoBLE.send((const char *)data,4);
	}
    }
    My_Decoder.DumpResults();
  }
}

void RFduinoBLE_onConnect()
{
  Serial.println( "onConnect" );
  clientConnected = 1;
}

void RFduinoBLE_onDisconnect()
{
  Serial.println( "onDisconnect" );
  clientConnected = 0;
}

void RFduinoBLE_onReceive(char *data, int len){
}

